// MSKGen.cpp: implementation of the CMSKGen class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "wscgen.h"
#include "MSKGen.h"
#include "math.h"

extern float BAUD;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define RMSCONSTANT 1.0
double SYMBOL_RATE = 45.45; //31.25		// 31.25 Symbols per Second bytheway
double Freq_Shift = 170.0/2.0; //(SYMBOL_RATE*4.0)	bytheway

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMSKGen::CMSKGen()
{
	m_2PI = 8.0 * atan(1.0);		//make constant 2 PI

}

CMSKGen::~CMSKGen()
{

}

/////////////////////////////////////////////////////////////
//Initialize the MSK generator with the sample frequncy Fs,
// the carrier frequency freq.
/////////////////////////////////////////////////////////////
void CMSKGen::InitMSK(INT Fs, double freq)
{
	SYMBOL_RATE = BAUD;
	SetMSKFreq( Fs, freq,0,0);	//set carrier frequency
	m_MSKSecPerSamp = 1.0/Fs;
	m_MSKTime = 0;
	m_MSKPeriodUpdate = 1.0/SYMBOL_RATE;

}

/////////////////////////////////////////////////////////////
// Call to change the psk31 carrier frequency
/////////////////////////////////////////////////////////////
void CMSKGen::SetMSKFreq(INT Fs, double freq, INT rate, INT range)
{
	m_MSKPhaseIncH = m_2PI * freq/(double)Fs;		//carrier frequency
	m_MSKPhaseIncL = m_2PI * freq/(double)Fs;		//carrier frequency
	m_MSKPhaseIncH += (m_2PI * Freq_Shift/(double)Fs);
	m_MSKPhaseIncL -= (m_2PI * Freq_Shift/(double)Fs);
}

/////////////////////////////////////////////////////////////
// generates n samples of MSK waveform in pData with average RMS value of 1.0
/////////////////////////////////////////////////////////////
int CMSKGen::CalcMSK(double *pData, INT n, int ch, int bitlen)
{
static double tc = 0.0;
INT i;
INT mod = -1;
int j = 0;
int count;
int input = ch;

	switch(bitlen) // Logic is inverted, need to add masks
	{
	case 5: ch = ch | 0xE0; break;
	case 6: ch = ch | 0xB0; break;
	case 7: ch = ch | 0x80; break;
	case 8: ch = ch;        break;
	default:
		break;
	}

	for( i=0; i<n; i++ )
	{
		m_MSKTime += m_MSKSecPerSamp;
		if(( m_MSKTime >= m_MSKPeriodUpdate ) && (input != 0))// && (j<bitlen))
		{
			count = i;
			switch(j) // bits are read from right
			{
			case 0:					mod = 1; break;  // Start pulse, send a mark tone
			case 1: if (ch & 0x01)	mod =-1; else mod = 1;  break;
			case 2: if (ch & 0x02)	mod =-1; else mod = 1;  break;
			case 3: if (ch & 0x04)	mod =-1; else mod = 1;  break;
			case 4: if (ch & 0x08)	mod =-1; else mod = 1;  break;
			case 5: if (ch & 0x10)	mod =-1; else mod = 1;  break;
			case 6: if (ch & 0x20)	mod =-1; else mod = 1;	break;
			case 7: if (ch & 0x40)	mod =-1; else mod = 1;	break;
			case 8: if (ch & 0x80)	mod =-1; else mod = 1;	break;
			default:
				mod = -1;
				break;
			}
			j++; 
			m_MSKTime -= m_MSKPeriodUpdate;	//keep time bounded
		}
		if(mod>0)
			tc = tc + (m_MSKPhaseIncH);	// increment radian phase count
		else
			tc = tc + (m_MSKPhaseIncL);	// increment radian phase count

		tc = fmod(tc,m_2PI);			//keep radian counter bounded


		pData[i] = RMSCONSTANT*( sin( tc ) + cos( tc ) );
	}
	return count;
}