// ProcThread.cpp: implementation of the CProcThread class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ProcThread.h"
#include "math.h"
//#include "..\codes.h"
#include "errorcodes.h"
#include "../defines.h" // Serial port configurations

extern char last_state;
extern char state;
extern int CalcMSK_TX_char[100];
extern bool CalcMSK_BuffEmpty;
/* Table of Baudot and AMTOR characters in ASCII order.*/
struct BaudotTable
{
			unsigned int state;
			char         ita2;  /* ita2 morray*/
			char         telex;    /* telex code */
			char         military; /* military code*/
};
extern struct BaudotTable table3[100];

BOOL m_fProcThreadQuit = true;			// signal to thread to quit


// Non Class routine for launching the worker thread

static UINT ThreadLauncher( CProcThread* pIO)
{
	pIO->ProcessLoop();
	return(4);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CProcThread::CProcThread()
{
	m_pProcThread = NULL;
	m_fProcThreadQuit = FALSE;
	m_pSoundDev = NULL;
//	m_pWaveDev = NULL;
//	m_pGenSrc = NULL;
//	m_pPSKSrc = NULL;
//	m_pSSTVSrc = NULL;
//	m_p400BPSKSrc = NULL;
//	m_pMSKSrc = NULL;
	m_pOutBuf = NULL;
	m_SNRatio = 20.0;
}

CProcThread::~CProcThread()
{

}

///////////////////////////////////////////////////////////////////////////
//  Stop process thread proceedure and wait for it to finish
//  Called by main program thread to signal data processing thread to stop
//	and then wait for it to stop before returning.
///////////////////////////////////////////////////////////////////////////
void CProcThread::StopThread()
{
	if( m_pProcThread != NULL )		//if thread is already running
	{
		m_fProcThreadQuit = TRUE;	// stop it
		::WaitForSingleObject(m_pProcThread->m_hThread,	// drawing thread handle
								INFINITE);
		delete m_pProcThread;
		m_pProcThread = NULL;
	}
}

///////////////////////////////////////////////////////////////////////////
//  Free resources and exit process thread
//   Called by process thread itself to shut itself off if not signaled by
//  main program thread.
///////////////////////////////////////////////////////////////////////////
void CProcThread::StopMyself()
{
//  if any resources are open, kill em.
	DeleteResources();
	m_pProcThread->m_bAutoDelete = TRUE;	//make thread clean up after
	m_pProcThread = NULL;  					// itself before it dies
//	m_pDlg->ProcessAbort();					//let main thread know I'm dead
	AfxEndThread(3); 
}

///////////////////////////////////////////////////////////////////////////
//   Called to start the thread running
///////////////////////////////////////////////////////////////////////////
void CProcThread::StartThread(void* pDlg )
{
	m_pDlg = (CRTTYAppDlg*)pDlg;				//save off pointer to parent class
	if( m_pProcThread == NULL)
	{
		m_fProcThreadQuit = FALSE;
		m_pProcThread = AfxBeginThread(
			(AFX_THREADPROC)ThreadLauncher,		//thread function
			(LPVOID)this,						// ptr to the class
			THREAD_PRIORITY_NORMAL, 			//give worker thread priority
//			THREAD_PRIORITY_HIGHEST,
			0,									// same stack size
			CREATE_SUSPENDED,					// don't let it start yet
			NULL );							// same security attributes
		m_pProcThread->m_bAutoDelete = FALSE;	// keep thread object around
		m_pProcThread->ResumeThread();	//let er rip
	}
}

///////////////////////////////////////////////////////////////////////////
//  Main I/O thread processing loop
///////////////////////////////////////////////////////////////////////////
void CProcThread::ProcessLoop()
{
WAVEFORMATEX Wfx;
INT BufSize;
DWORD dwOutBlockSize;
DWORD SampleLimit;
BOOL fFinished = FALSE;
BOOL result;
//DWORD SamplesSaved;
UINT ErrorCode;
double databits,stopbits,bits;
extern int DATABITS;
extern float BAUD;

	Wfx.wFormatTag = WAVE_FORMAT_PCM;
	Wfx.nChannels = 1;					//don't use stereo
	Wfx.nSamplesPerSec = m_pDlg->m_SampsPerSec;
	Wfx.nAvgBytesPerSec = m_pDlg->m_SampsPerSec*1*2;	//is 16 bit data
	Wfx.nBlockAlign = 1*2;				//number bytes per sample
	Wfx.wBitsPerSample = 16;
	Wfx.cbSize = 0;
//	BufSize = 4096;				// buffer size
//	BufSize = 1310; // bytheway RTTY loop

	switch(DATABITS)
	{
	case d5: databits = 5.0; stopbits = 1.5; break;
	case d6: databits = 6.0; stopbits = 2.0; break;
	case d7: databits = 7.0; stopbits = 2.0; break;
	case d8: databits = 8.0; stopbits = 2.0; break;
	default:
		databits = 5.0;
		stopbits = 2.0; 
		break;
	}
	bits = 1.0 + databits + stopbits;

	BufSize = (int) (Wfx.nSamplesPerSec*(1.0/BAUD)*bits);

	CalcSNR();


	switch(m_pDlg->m_SampsPerSec)	//determine soundcard buffersizes
	{							//(need a couple hundred milliseconds worth)
		case 8000:
		case 11025:
			dwOutBlockSize = BufSize;	//soundcard buffer size
			break;
		case 22050:
			dwOutBlockSize = BufSize*2;	//soundcard buffer size
			break;
		case 44100:
			dwOutBlockSize = BufSize*4;	//soundcard buffer size
			break;
	}


//allocate some buffers for signals
	if( (m_pOutBuf = new double[sizeof(double)*BufSize]) == NULL )
	{
		AfxMessageBox( _T("Output Memory Error") , MB_ICONSTOP );
		StopMyself();
	}

	SampleLimit = m_pDlg->m_TimeLimit * m_pDlg->m_SampsPerSec;

	// Determine the Sound Card Identifier Number
	#define BUTTON_UNCHECKED 0
	#define BUTTON_CHECKED 1
	char strcard[30];
	int cardoption;
	m_pDlg->m_ComboSoundCard.GetWindowText(strcard,sizeof(strcard));
/*	Options:
		No Sound
		-1 Auto
		1
		2
		3
		4
*/
	cardoption = atoi(strcard);
	switch(cardoption)
	{
	case -1: cardoption = -1; break; // Default Auto Select
	case  1: cardoption =  0; break; // Card 0
	case  2: cardoption =  1; break; // Card 1
	case  3: cardoption =  2; break; // Card 2
	case  4: cardoption =  3; break; // Card 3
	default: cardoption = 10;
		m_pDlg->m_CheckNoSoundCard.SetCheck(0);
		break; // it will throw out the option
	}

//instantiate either a soundcard or wavefile object for output.
	m_pSoundDev = new CSound;
	ErrorCode = m_pSoundDev->OutOpen(  &Wfx, dwOutBlockSize, 0,cardoption);
	if( ErrorCode != NO_ERROR )
	{
		DisplayError( ErrorCode );
		m_pDlg->CmRX();
		return;
//		StopMyself();
	}
	CClientDC dc(m_pDlg);				// get a screen device context for the
									// SSTV image generator functions

	m_pGenSrc = new CGenerators;		// use for sin and noise generation


// Initialize the correct generator that is to be used

	m_pMSKSrc = new CMSKGen;
	m_pMSKSrc->InitMSK( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq);

//
///////////////   M a i n    P r o c e s s i n g   L o o p   ///////////////	
//				
	
//	bool toggle = false;
//	static int character= 65;
	extern CRITICAL_SECTION vCriticalSection;
	int i;
	int send_char;
	extern bool CalcMSK_BuffEmpty;

	while( !m_fProcThreadQuit)
	{

		//Generate Bufsize data samples
		EnterCriticalSection(&vCriticalSection);
		m_pMSKSrc->SetMSKFreq(m_pDlg->m_SampsPerSec,
							(double)m_pDlg->m_SinFreq,
							m_pDlg->m_SweepRate,
							m_pDlg->m_SweepRange );
		send_char = CalcMSK_TX_char[1];
		for (i=1;i<CalcMSK_TX_char[0]; i++) // shift down one
			CalcMSK_TX_char[i] = CalcMSK_TX_char[i+1];
		CalcMSK_TX_char[0]--; // one less character in buffer
		if (CalcMSK_TX_char[0] <= 0)
		{
			CalcMSK_TX_char[0] = 0; //set counter to zero
			CalcMSK_TX_char[1] = 0; // set first character to zero
			CalcMSK_BuffEmpty = true;
		}
		else
			CalcMSK_BuffEmpty = false;
		int count = m_pMSKSrc->CalcMSK(m_pOutBuf, BufSize, send_char,(int)databits);
//		dwOutBlockSize = count;



		// Generate BufSize Gaussian noise samples and add to signal with S/N values
		m_pGenSrc->AddBWLimitedNoise(BufSize, m_pOutBuf, m_SignalAmplitude,
									m_NoiseAmplitude,m_pDlg->m_BWLimit );
		LeaveCriticalSection(&vCriticalSection);
		//output data to soundcard object or wave file object
		if( (result = m_pSoundDev->OutWrite( m_pOutBuf, BufSize )) != BufSize)
		{
			if( result != 0)
			{
				ErrorCode = m_pSoundDev->GetError();
				DisplayError( ErrorCode );
			}
			m_pSoundDev->OutClose();
			StopMyself();
		}

		if(fFinished)
		{
			StopMyself();
		}

	}
	m_pSoundDev->Flush();
// here if main thread has signaled this thread to shut down
	DeleteResources();		//kill all resources

}


///////////////////////////////////////////////////////////////////////////
//   Called to delete any allocated resources in this class
///////////////////////////////////////////////////////////////////////////
void CProcThread::DeleteResources()
{
//  if any object pointer is active, kill it.
	if(	m_pSoundDev)
	{
		m_pSoundDev->OutClose();
		delete m_pSoundDev;
		m_pSoundDev = NULL;
	}

	if(	m_pMSKSrc)
	{
		delete m_pMSKSrc;
		m_pMSKSrc = NULL;
	}

	if(m_pOutBuf)
	{
		delete m_pOutBuf;	
		m_pOutBuf = NULL;
	}
}

/////////////////////////////////////////////////////////////////////
// Simple minded error message display for various sound/wave errors
// Should be put in string table.
void CProcThread::DisplayError( INT ErrorCode )
{
CString Msg;
	switch( ErrorCode )
	{
		case NO_ERRORS:
			Msg = _T("");
			break;
		case MEMORY_ERROR:
			Msg = _T("Memory Allocation error");
			break;
		case WAVIN_ERR_OPEN:
			Msg = _T("Can't open wave file for input");
			break;
		case WAVIN_ERR_NOTWAVE:
			Msg = _T("File is not a RIFF wave type");
			break;
		case WAVIN_ERR_INVALID:
			Msg = _T("Invalid wave file");
			break;
		case WAVIN_ERR_NODATA:
			Msg = _T("No data in file");
			break;
		case WAVIN_ERR_NOTSUPPORTED:
			Msg = _T("Not a supported data type");
			break;
		case WAVIN_ERR_READING:
			Msg = _T("Error reading data from file");
			break;
		case WAVIN_ERR_NOTOPEN:
			Msg = _T("Tried to read and file is not open");
			break;
		case WAVOUT_ERR_OPEN:
			Msg = _T("Can't open wave file for input");
			break;
		case WAVOUT_ERR_WRITING:
			Msg = _T("Error writing to wave file");
			break;
		case WAVOUT_ERR_NOTOPEN:
			Msg = _T("Tried to write and file is not open");
			break;
		case SOUNDIN_ERR_NOTOPEN:
			Msg = _T("Tried to read and soundcard is not open");
			break;
		case SOUNDIN_ERR_OVERFLOW:
			Msg = _T("Input buffers overflowed");
			break;
		case SOUNDIN_ERR_TIMEOUT:
			Msg = _T("Timed out waiting for input buffers");
			break;
		case SOUNDOUT_ERR_NOTOPEN:
			Msg = _T("Tried to write and soundcard is not open");
			break;
		case SOUNDOUT_ERR_UNDERFLOW:
			Msg = _T("Output buffers Underflowed");
			break;
		case SOUNDOUT_ERR_TIMEOUT:
			Msg = _T("Timed out waiting for output buffers");
			break;
//
		case MMSYSERR_BADDEVICEID:
			Msg = _T("Device ID out of range");
			break;
		case MMSYSERR_NOTENABLED:
			Msg = _T("Driver failed enable");
			break;
		case MMSYSERR_ALLOCATED:
			Msg = _T("Device already allocated");
			break;
		case MMSYSERR_INVALHANDLE:
			Msg = _T("Device handle is invalid");
			break;
		case MMSYSERR_NODRIVER:
			Msg = _T("No device driver present");
			break;
		case MMSYSERR_NOMEM:
			Msg = _T("mmsystem memory allocation error");
			break;
		case MMSYSERR_NOTSUPPORTED:
			Msg = _T("Function isn't supported");
			break;
		case MMSYSERR_BADERRNUM:
			Msg = _T("Error value out of range");
			break;
		case MMSYSERR_INVALFLAG: 
			Msg = _T("Invalid flag passed");
			break;
		case MMSYSERR_INVALPARAM: 
			Msg = _T("Invalid parameter passed");
			break;
		case MMSYSERR_HANDLEBUSY:
			Msg = _T("Handle being used");
			break;
		default:			// unknown error code
			Msg = _T("??? Maybe Input being used at\n a different sample rate");
			break;
	}
	if( Msg != _T("") )
		AfxMessageBox( Msg, MB_ICONERROR );
}

///////////////////////////////////////////////////////////////////
//calculates RMS value of buf[] samples over RMS_SIZE*numsamps
//  called once for each buffer to be measured
///////////////////////////////////////////////////////////////////
double CProcThread::CalcRMS(double* buf, INT bufsize)
{
#define RMS_SIZE 10
static double sum = 0.0;
static int cnt = 0;
static double rms = 0.0;
	for( INT i = 0; i<bufsize; i++)
		sum = sum + ( buf[i]*buf[i] );
	if( ++cnt >= RMS_SIZE)
	{
		rms = sqrt(sum/(cnt*bufsize));
		sum = 0.0;
		cnt = 0;
	}
	return rms;
}

/////////  local defines
#define MAX_SNR 60.0
#define MIN_SNR -40.0
#define MIN_FREQ 10
#define BPSK_RMS_CONSTANT 1.1247// Fudge factor to make average BPSK signal RMS = 1
#define BPSK_CONSTANT (BPSK_RMS_CONSTANT/1.414)
#define RMS_AMPLITUDE 4739.0	//RMS amplitude out of 32768
								// pick so that worst case Gaussian
								// plus sin wave will not overflow soundcard
								// 32768 = RMS*5.5 + 1.414*RMS
								// RMS = 4739;

void CProcThread::CalcSNR()
{
	extern CRITICAL_SECTION vCriticalSection;
	m_SNRatio = 200.0;

	if( m_SNRatio > MAX_SNR )
		m_SNRatio = MAX_SNR;
	if( m_SNRatio < MIN_SNR )
		m_SNRatio = MIN_SNR;
//	UpdateData(FALSE);		// update edit box to new value
	EnterCriticalSection(&vCriticalSection);
	if( m_SNRatio >= 0.0 )
	{
		m_SignalAmplitude = RMS_AMPLITUDE;
		m_NoiseAmplitude = RMS_AMPLITUDE/pow(10 , m_SNRatio/20.0);
	}
	else
	{
		m_NoiseAmplitude = RMS_AMPLITUDE;
		m_SignalAmplitude = RMS_AMPLITUDE*pow(10 , m_SNRatio/20.0);
	}
	LeaveCriticalSection(&vCriticalSection);

}

void CProcThread::MSK_xmit(int ch)
{
	extern CRITICAL_SECTION vCriticalSection;	// use for keeping threads

	EnterCriticalSection(&vCriticalSection);
	  CalcMSK_TX_char[0]++;
	  if (CalcMSK_TX_char[0] > 99)
		  CalcMSK_TX_char[0] = 99;
	  CalcMSK_TX_char[CalcMSK_TX_char[0]] = ch;
	LeaveCriticalSection(&vCriticalSection);
}


